/*********************************************************************
 * base.h
 *********************************************************************/

#ifndef _BASE_H_
#define _BASE_H_

#ifndef uchar
#define uchar unsigned char
#endif

#ifndef schar
#define schar signed char
#endif

#ifndef uint
#define uint unsigned int
#endif

#ifndef ushort
#define ushort unsigned short
#endif

#ifndef ulong
#define ulong unsigned long
#endif
#endif
